const Expr = @import("./Expr.zig");

const Stmt = @import("./Stmt.zig");

const app = @import("../app.zig");

const coral = @import("coral");

const file = @import("../file.zig");

const kym = @import("../kym.zig");

const tree = @import("./tree.zig");

name: *kym.RuntimeRef,
arity: u8,
opcodes: OpcodeList,
lines: LineList,
cursor: usize,
constants: ConstList,
bindings: []?*kym.RuntimeRef,

const Builtin = enum {
	import,
	print,
	vec2,
	vec3,
};

const Compiler = struct {
	chunk: *Self,
	env: *kym.RuntimeEnv,

	fn compile_argument(self: *const Compiler, environment: *const tree.Environment, initial_argument: ?*const Expr) kym.RuntimeError!u8 {
		// TODO: Exceeding 255 arguments will make the VM crash.
		var maybe_argument = initial_argument;
		var argument_count = @as(u8, 0);

		while (maybe_argument) |argument| {
			try self.compile_expression(environment, argument, null);

			maybe_argument = argument.next;
			argument_count += 1;
		}

		return argument_count;
	}

	fn compile_expression(self: *const Compiler, environment: *const tree.Environment, expression: *const Expr, name: ?[]const coral.io.Byte) kym.RuntimeError!void {
		const number_format = coral.utf8.DecimalFormat{
			.delimiter = "_",
			.positive_prefix = .none,
		};

		switch (expression.kind) {
			.nil_literal => try self.chunk.write(expression.line, .push_nil),
			.true_literal => try self.chunk.write(expression.line, .push_true),
			.false_literal => try self.chunk.write(expression.line, .push_false),

			.number_literal => |literal| {
				for (literal) |codepoint| {
					if (codepoint == '.') {
						return self.chunk.write(expression.line, .{
							.push_const = try self.declare_float(number_format.parse(literal, kym.Float) orelse unreachable),
						});
					}
				}

				try self.chunk.write(expression.line, .{
					.push_const = try self.declare_fixed(number_format.parse(literal, kym.Fixed) orelse unreachable),
				});
			},

			.string_literal => |literal| {
				try self.chunk.write(expression.line, .{.push_const = try self.declare_string(literal)});
			},

			.symbol_literal => |literal| {
				try self.chunk.write(expression.line, .{.push_const = try self.declare_symbol(literal)});
			},

			.table_construct => |table_construct| {
				var table_entry = table_construct.entry;
				var field_count = @as(u32, 0);

				while (table_entry) |entry| : (table_entry = entry.next) {
					try self.compile_expression(environment, entry, null);

					if (entry.kind != .key_value) {
						try self.chunk.write(expression.line, .push_top);
					}

					field_count += 1;
				}

				try self.chunk.write(expression.line, .{.push_table = field_count});
			},

			.key_value => |key_value| {
				try self.compile_expression(environment, key_value.value, null);
				try self.compile_expression(environment, key_value.key, null);
			},

			.lambda_construct => |lambda_construct| {
				var chunk = try Self.make(self.env, name orelse "<lambda>", lambda_construct.environment);

				errdefer chunk.free(self.env);

				if (lambda_construct.environment.capture_count == 0) {
					try self.chunk.write(expression.line, .{.push_const = try self.declare_chunk(chunk)});
				} else {
					const lambda_captures = lambda_construct.environment.get_captures();
					var index = lambda_captures.len;

					while (index != 0) {
						index -= 1;

						try self.chunk.write(expression.line, switch (lambda_captures[index]) {
							.declaration_index => |declaration_index| .{.push_local = declaration_index},
							.capture_index => |capture_index| .{.push_binding = capture_index},
						});
					}

					try self.chunk.write(expression.line, .{.push_const = try self.declare_chunk(chunk)});
					try self.chunk.write(expression.line, .{.bind = lambda_construct.environment.capture_count});
				}
			},

			.binary_op => |binary_op| {
				try self.compile_expression(environment, binary_op.lhs_operand, null);
				try self.compile_expression(environment, binary_op.rhs_operand, null);

				try self.chunk.write(expression.line, switch (binary_op.operation) {
					.addition => .add,
					.subtraction => .sub,
					.multiplication => .mul,
					.divsion => .div,
					.greater_equals_comparison => .cge,
					.greater_than_comparison => .cgt,
					.equals_comparison => .eql,
					.less_than_comparison => .clt,
					.less_equals_comparison => .cle,
				});
			},

			.unary_op => |unary_op| {
				try self.compile_expression(environment, unary_op.operand, null);

				try self.chunk.write(expression.line, switch (unary_op.operation) {
					.boolean_negation => .not,
					.numeric_negation => .neg,
				});
			},

			.invoke => |invoke| {
				const argument_count = try self.compile_argument(environment, invoke.argument);

				try self.compile_expression(environment, invoke.object, null);
				try self.chunk.write(expression.line, .{.call = argument_count});
			},

			.group => |group| try self.compile_expression(environment, group, null),
			.import_builtin => try self.chunk.write(expression.line, .{.push_builtin = .import}),
			.print_builtin => try self.chunk.write(expression.line, .{.push_builtin = .print}),
			.vec2_builtin => try self.chunk.write(expression.line, .{.push_builtin = .vec2}),
			.vec3_builtin => try self.chunk.write(expression.line, .{.push_builtin = .vec3}),

			.declaration_get => |declaration_get| {
				if (get_local_index(environment, declaration_get.declaration)) |index| {
					if (is_declaration_boxed(declaration_get.declaration)) {
						try self.chunk.write(expression.line, .{.push_local = index});
						try self.chunk.write(expression.line, .get_box);
					} else {
						try self.chunk.write(expression.line, .{.push_local = index});
					}

					return;
				}

				if (try self.get_binding_index(environment, declaration_get.declaration)) |index| {
					try self.chunk.write(expression.line, .{.push_binding = index});

					if (is_declaration_boxed(declaration_get.declaration)) {
						try self.chunk.write(expression.line, .get_box);
					}

					return;
				}

				return self.env.raise(error.IllegalState, "local out of scope", .{});
			},

			.declaration_set => |declaration_set| {
				if (get_local_index(environment, declaration_set.declaration)) |index| {
					try self.compile_expression(environment, declaration_set.assign, null);

					if (is_declaration_boxed(declaration_set.declaration)) {
						try self.chunk.write(expression.line, .{.push_local = index});
						try self.chunk.write(expression.line, .set_box);
					} else {
						try self.chunk.write(expression.line, .{.set_local = index});
					}

					return;
				}

				if (try self.get_binding_index(environment, declaration_set.declaration)) |index| {
					try self.compile_expression(environment, declaration_set.assign, null);
					try self.chunk.write(expression.line, .{.push_binding = index});

					if (is_declaration_boxed(declaration_set.declaration)) {
						try self.chunk.write(expression.line, .set_box);
					}

					return;
				}

				return self.env.raise(error.IllegalState, "local out of scope", .{});
			},

			.field_get => |field_get| {
				try self.compile_expression(environment, field_get.object, null);
				try self.chunk.write(expression.line, .{.push_const = try self.declare_symbol(field_get.identifier)});
				try self.chunk.write(expression.line, .get_dynamic);
			},

			.field_set => |field_set| {
				try self.compile_expression(environment, field_set.object, null);
				try self.chunk.write(expression.line, .{.push_const = try self.declare_symbol(field_set.identifier)});
				try self.compile_expression(environment, field_set.assign, null);
				try self.chunk.write(expression.line, .set_dynamic);
			},

			.subscript_get => |subscript_get| {
				try self.compile_expression(environment, subscript_get.object, null);
				try self.compile_expression(environment, subscript_get.index, null);
				try self.chunk.write(expression.line, .get_dynamic);
			},

			.subscript_set => |subscript_set| {
				try self.compile_expression(environment, subscript_set.object, null);
				try self.compile_expression(environment, subscript_set.index, null);
				try self.compile_expression(environment, subscript_set.assign, null);
				try self.chunk.write(expression.line, .set_dynamic);
			},
		}
	}

	pub fn compile_environment(self: *const Compiler, environment: *const tree.Environment) kym.RuntimeError!void {
		if (environment.statement) |statement| {
			const last_statement = try self.compile_statement(environment, statement);

			if (last_statement.kind != .@"return") {
				try self.chunk.write(last_statement.line, .push_nil);
			}
		}
	}

	fn compile_statement(self: *const Compiler, environment: *const tree.Environment, initial_statement: *const Stmt) kym.RuntimeError!*const Stmt {
		var current_statement = initial_statement;

		while (true) {
			switch (current_statement.kind) {
				.@"return" => |@"return"| {
					if (@"return".returned_expression) |expression| {
						try self.compile_expression(environment, expression, null);
					} else {
						try self.chunk.write(current_statement.line, .push_nil);
					}

					// TODO: Omit ret calls at ends of chunk.
					try self.chunk.write(current_statement.line, .ret);
				},

				.@"while" => |@"while"| {
					try self.compile_expression(environment, @"while".loop_expression, null);
					try self.chunk.write(current_statement.line, .{.jf = 0});

					const origin_index = @as(u32, @intCast(self.chunk.opcodes.values.len - 1));

					_ = try self.compile_statement(environment, @"while".loop);
					self.chunk.opcodes.values[origin_index].jf = @intCast(self.chunk.opcodes.values.len - 1);

					try self.compile_expression(environment, @"while".loop_expression, null);
					try self.chunk.write(current_statement.line, .{.jt = origin_index});
				},

				.@"if" => |@"if"| {
					try self.compile_expression(environment, @"if".then_expression, null);
					try self.chunk.write(current_statement.line, .{.jf = 0});

					const origin_index = @as(u32, @intCast(self.chunk.opcodes.values.len - 1));

					_ = try self.compile_statement(environment, @"if".@"then");
					self.chunk.opcodes.values[origin_index].jf = @intCast(self.chunk.opcodes.values.len - 1);

					if (@"if".@"else") |@"else"| {
						_ = try self.compile_statement(environment, @"else");
					}
				},

				.declare => |declare| {
					try self.compile_expression(environment, declare.initial_expression, declare.declaration.identifier);

					if (is_declaration_boxed(declare.declaration)) {
						try self.chunk.write(current_statement.line, .push_boxed);
					}
				},

				.top_expression => |top_expression| {
					try self.compile_expression(environment, top_expression, null);

					if (top_expression.kind == .invoke) {
						try self.chunk.write(current_statement.line, .pop);
					}
				},
			}

			current_statement = current_statement.next orelse return current_statement;
		}
	}

	const constants_max = @as(usize, coral.math.max_int(@typeInfo(u16).Int));

	fn declare_chunk(self: *const Compiler, chunk: Self) kym.RuntimeError!u16 {
		if (self.chunk.constants.values.len == coral.math.max_int(@typeInfo(u16).Int)) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = @as(usize, coral.math.max_int(@typeInfo(u16).Int)),
			});
		}

		const constant = try self.env.new_dynamic(coral.io.bytes_of(&chunk), typeinfo);

		errdefer self.env.discard(constant);

		try self.chunk.constants.push_one(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn declare_fixed(self: *const Compiler, fixed: kym.Fixed) kym.RuntimeError!u16 {
		if (self.chunk.constants.values.len == constants_max) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = constants_max,
			});
		}

		const constant = try self.env.new_fixed(fixed);

		errdefer self.env.discard(constant);

		try self.chunk.constants.push_one(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn declare_float(self: *const Compiler, float: kym.Float) kym.RuntimeError!u16 {
		if (self.chunk.constants.values.len == constants_max) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = constants_max,
			});
		}

		const constant = try self.env.new_float(float);

		errdefer self.env.discard(constant);

		try self.chunk.constants.push_one(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn declare_string(self: *const Compiler, string: []const coral.io.Byte) kym.RuntimeError!u16 {
		if (self.chunk.constants.values.len == constants_max) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = constants_max,
			});
		}

		const constant = try self.env.new_string(string);

		errdefer self.env.discard(constant);

		try self.chunk.constants.push_one(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn declare_symbol(self: *const Compiler, symbol: []const coral.io.Byte) kym.RuntimeError!u16 {
		if (self.chunk.constants.values.len == constants_max) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = constants_max,
			});
		}

		const constant = try self.env.new_symbol(symbol);

		errdefer self.env.discard(constant);

		try self.chunk.constants.push_one(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn get_binding_index(self: *const Compiler, environment: *const tree.Environment, declaration: *const tree.Declaration) kym.RuntimeError!?u8 {
		var binding_index = @as(u8, 0);

		while (binding_index < environment.capture_count) : (binding_index += 1) {
			var capture = &environment.captures[binding_index];
			var target_environment = environment.enclosing orelse return null;

			while (capture.* == .capture_index) {
				capture = &target_environment.captures[capture.capture_index];
				target_environment = target_environment.enclosing orelse return null;
			}

			try kym.assert(self.env, capture.* == .declaration_index);

			if (&target_environment.declarations[capture.declaration_index] == declaration) {
				return binding_index;
			}
		}

		return null;
	}

	fn get_local_index(environment: *const tree.Environment, declaration: *const tree.Declaration) ?u8 {
		var remaining = environment.declaration_count;

		while (remaining != 0) {
			remaining -= 1;

			if (&environment.declarations[remaining] == declaration) {
				return remaining;
			}
		}

		return null;
	}

	fn is_declaration_boxed(declaration: *const tree.Declaration) bool {
		return declaration.is.captured and !declaration.is.readonly;
	}
};

const ConstList = coral.list.Stack(*kym.RuntimeRef);

const LineList = coral.list.Stack(u32);

pub const Opcode = union (enum) {
	ret,
	pop,
	push_nil,
	push_true,
	push_false,
	push_const: u16,
	push_local: u8,
	push_top,
	push_table: u32,
	push_builtin: Builtin,
	push_binding: u8,
	push_boxed,
	set_local: u8,
	get_dynamic,
	set_dynamic,
	get_box,
	set_box,
	call: u8,
	bind: u8,

	not,
	neg,

	add,
	sub,
	mul,
	div,

	eql,
	cgt,
	clt,
	cge,
	cle,

	jt: u32,
	jf: u32,
};

const OpcodeList = coral.list.Stack(Opcode);

const Self = @This();

pub fn dump(chunk: Self, env: *kym.RuntimeEnv) kym.RuntimeError!*kym.RuntimeRef {
	var opcode_cursor = @as(u32, 0);
	var buffer = coral.list.ByteStack.make(env.allocator);

	defer buffer.free();

	const writer = coral.list.stack_as_writer(&buffer);

	_ = coral.utf8.print_string(writer, "\n");

	while (opcode_cursor < chunk.opcodes.values.len) : (opcode_cursor += 1) {
		_ = coral.utf8.print_formatted(writer, "[{instruction}]: ", .{.instruction = opcode_cursor});

		_ = switch (chunk.opcodes.values[opcode_cursor]) {
			.ret => coral.utf8.print_string(writer, "ret\n"),
			.pop => coral.utf8.print_string(writer, "pop\n"),
			.push_nil => coral.utf8.print_string(writer, "push nil\n"),
			.push_true => coral.utf8.print_string(writer, "push true\n"),
			.push_false => coral.utf8.print_string(writer, "push false\n"),

			.push_const => |push_const| print: {
				try kym.assert(env, push_const < chunk.constants.values.len);

				const string_ref = try env.to_string(chunk.constants.values[push_const]);

				defer env.discard(string_ref);

				const string = string_ref.as_string();

				coral.debug.assert(string != null);

				break: print coral.utf8.print_formatted(writer, "push const ({value})\n", .{.value = string.?});
			},

			.push_local => |push_local| coral.utf8.print_formatted(writer, "push local ({local})\n", .{
				.local = push_local,
			}),

			.push_top => coral.utf8.print_string(writer, "push top\n"),

			.push_table => |push_table| coral.utf8.print_formatted(writer, "push table ({count})\n", .{
				.count = push_table,
			}),

			.push_boxed => coral.utf8.print_string(writer, "push boxed\n"),

			.push_binding => |push_binding| coral.utf8.print_formatted(writer, "push binding ({binding})\n", .{
				.binding = push_binding,
			}),

			.push_builtin => |push_builtin| coral.utf8.print_formatted(writer, "push builtin ({builtin})\n", .{
				.builtin = switch (push_builtin) {
					.import => "import",
					.print => "print",
					.vec2 => "vec2",
					.vec3 => "vec3",
				},
			}),

			.bind => |bind| coral.utf8.print_formatted(writer, "bind ({count})\n", .{
				.count = bind,
			}),

			.set_local => |local_set| coral.utf8.print_formatted(writer, "set local ({local})\n", .{
				.local = local_set,
			}),

			.get_box => coral.utf8.print_string(writer, "get box\n"),
			.set_box => coral.utf8.print_string(writer, "set box\n"),
			.get_dynamic => coral.utf8.print_string(writer, "get dynamic\n"),
			.set_dynamic => coral.utf8.print_string(writer, "set dynamic\n"),
			.call => |call| coral.utf8.print_formatted(writer, "call ({count})\n", .{.count = call}),
			.not => coral.utf8.print_string(writer, "not\n"),
			.neg => coral.utf8.print_string(writer, "neg\n"),
			.add => coral.utf8.print_string(writer, "add\n"),
			.sub => coral.utf8.print_string(writer, "sub\n"),
			.mul => coral.utf8.print_string(writer, "mul\n"),
			.div => coral.utf8.print_string(writer, "div\n"),
			.eql => coral.utf8.print_string(writer, "eql\n"),
			.cgt => coral.utf8.print_string(writer, "cgt\n"),
			.clt => coral.utf8.print_string(writer, "clt\n"),
			.cge => coral.utf8.print_string(writer, "cge\n"),
			.cle => coral.utf8.print_string(writer, "cle\n"),
			.jf => |jf| coral.utf8.print_formatted(writer, "jf ({instruction})\n", .{.instruction = jf}),
			.jt => |jt| coral.utf8.print_formatted(writer, "jt ({instruction})\n", .{.instruction = jt}),
		};
	}

	return env.new_string(buffer.values);
}

pub fn execute(self: *Self, env: *kym.RuntimeEnv, frame: *const kym.Frame) kym.RuntimeError!?*kym.RuntimeRef {
	self.cursor = 0;

	while (self.cursor < self.opcodes.values.len) : (self.cursor += 1) {
		switch (self.opcodes.values[self.cursor]) {
			.ret => break,

			.pop => {
				if (try env.pop_local()) |ref| {
					env.discard(ref);
				}
			},

			.push_nil => try env.locals.push_one(null),
			.push_true => try env.locals.push_one(try env.new_boolean(true)),
			.push_false => try env.locals.push_one(try env.new_boolean(false)),

			.push_const => |push_const| {
				try kym.assert(env, push_const < self.constants.values.len);
				try env.locals.push_one(self.constants.values[push_const].acquire());
			},

			.push_local => |push_local| {
				try kym.assert(env, push_local < (env.locals.values.len - frame.locals_top));

				if (env.locals.values[frame.locals_top + push_local]) |local| {
					try env.locals.push_one(local.acquire());
				} else {
					try env.locals.push_one(null);
				}
			},

			.push_top => {
				const frame_locals = env.locals.values[frame.locals_top ..];

				try kym.assert(env, frame_locals.len != 0);

				if (frame_locals[frame_locals.len - 1]) |local| {
					try env.locals.push_one(local.acquire());
				} else {
					try env.locals.push_one(null);
				}
			},

			.push_table => |push_table| {
				const table = try env.new_table();

				errdefer env.discard(table);

				{
					var popped = @as(usize, 0);

					while (popped < push_table) : (popped += 1) {
						const index = try env.expect(try env.pop_local());

						defer env.discard(index);

						if (try env.pop_local()) |value| {
							defer env.discard(value);

							try env.set(table, index, value);
						}
					}
				}

				try env.locals.push_one(table);
			},

			.push_boxed => {
				const value = try env.pop_local();

				defer {
					if (value) |ref| {
						env.discard(ref);
					}
				}

				const boxable = try env.new_boxed(value);

				errdefer env.discard(boxable);

				try env.locals.push_one(boxable);
			},

			.push_binding => |push_binding| {
				try kym.assert(env, push_binding <= self.bindings.len);
				try env.locals.push_one(if (self.bindings[push_binding]) |value| value.acquire() else null);
			},

			.bind => |bind| {
				const callable = try env.expect(try env.pop_local());

				errdefer env.discard(callable);

				const chunk = @as(*Self, @ptrCast(@alignCast(try env.unwrap_dynamic(callable, typeinfo))));

				chunk.bindings = try coral.io.allocate_many(env.allocator, bind, @as(?*kym.RuntimeRef, null));

				for (0 .. bind) |index| {
					const value = try env.pop_local();

					errdefer {
						if (value) |ref| {
							env.discard(ref);
						}
					}

					const binding = &chunk.bindings[index];

					if (binding.*) |*existing_binding| {
						env.discard(existing_binding.*);
					}

					binding.* = value;
				}

				try env.locals.push_one(callable);
			},

			.push_builtin => |push_builtin| {
				const builtin_syscall = try env.new_syscall(switch (push_builtin) {
					.import => syscall_import,
					.print => syscall_print,
					.vec2 => syscall_vec2,
					.vec3 => syscall_vec3,
				});

				errdefer env.discard(builtin_syscall);

				try env.locals.push_one(builtin_syscall);
			},

			.set_local => |local_set| {
				const local = &env.locals.values[frame.locals_top + local_set];

				if (local.*) |previous_local| {
					env.discard(previous_local);
				}

				local.* = try env.pop_local();
			},

			.get_box => {
				const box = try env.expect(try env.pop_local());

				defer env.discard(box);

				if (try env.get_boxed(box)) |unboxed| {
					errdefer env.discard(unboxed);

					try env.locals.push_one(unboxed);
				} else {
					try env.locals.push_one(null);
				}
			},

			.set_box => {
				const box = try env.expect(try env.pop_local());

				defer env.discard(box);

				const value = try env.expect(try env.pop_local());

				errdefer env.discard(value);

				try env.set_boxed(box, value);
			},

			.get_dynamic => {
				const index = try env.expect(try env.pop_local());

				defer env.discard(index);

				const indexable = try env.expect(try env.pop_local());

				defer env.discard(indexable);

				const value = try env.get(indexable, index);

				errdefer {
					if (value) |ref| {
						env.discard(ref);
					}
				}

				try env.locals.push_one(value);
			},

			.set_dynamic => {
				const value = try env.pop_local();

				defer {
					if (value) |ref| {
						env.discard(ref);
					}
				}

				const index = try env.expect(try env.pop_local());

				defer env.discard(index);

				const indexable = try env.expect(try env.pop_local());

				defer env.discard(indexable);

				try env.set(indexable, index, value);
			},

			.call => |call| {
				const result = call: {
					const callable = try env.expect(try env.pop_local());

					defer env.discard(callable);

					const call_frame = try env.push_frame(callable, call);

					defer env.pop_frame();

					break: call try env.call_frame(&call_frame);
				};

				errdefer {
					if (result) |ref| {
						env.discard(ref);
					}
				}

				try env.locals.push_one(result);
			},

			.not => {
				if (try env.pop_local()) |value| {
					defer env.discard(value);

					try env.locals.push_one(try env.new_boolean(!value.is_truthy()));
				} else {
					try env.locals.push_one(try env.new_boolean(true));
				}
			},

			.neg => {
				const value = try env.expect(try env.pop_local());

				defer env.discard(value);

				try env.locals.push_one(try env.neg(value));
			},

			.add => {
				const rhs = try env.expect(try env.pop_local());

				defer env.discard(rhs);

				const lhs = try env.expect(try env.pop_local());

				defer env.discard(lhs);

				try env.locals.push_one(try env.add(lhs, rhs));
			},

			.sub => {
				const rhs = try env.expect(try env.pop_local());

				defer env.discard(rhs);

				const lhs = try env.expect(try env.pop_local());

				defer env.discard(lhs);

				try env.locals.push_one(try env.sub(lhs, rhs));
			},

			.mul => {
				const rhs = try env.expect(try env.pop_local());

				defer env.discard(rhs);

				const lhs = try env.expect(try env.pop_local());

				defer env.discard(lhs);

				try env.locals.push_one(try env.mul(lhs, rhs));
			},

			.div => {
				const rhs = try env.expect(try env.pop_local());

				defer env.discard(rhs);

				const lhs = try env.expect(try env.pop_local());

				defer env.discard(lhs);

				try env.locals.push_one(try env.div(lhs, rhs));
			},

			.eql => {
				if (try env.pop_local()) |rhs| {
					env.discard(rhs);

					if (try env.pop_local()) |lhs| {
						env.discard(lhs);

						try env.locals.push_one(try env.new_boolean(lhs.equals(rhs)));
					} else {
						try env.locals.push_one(try env.new_boolean(false));
					}
				} else {
					if (try env.pop_local()) |lhs| {
						env.discard(lhs);

						try env.locals.push_one(try env.new_boolean(false));
					} else {
						try env.locals.push_one(try env.new_boolean(true));
					}
				}
			},

			.cgt => {
				const rhs = try env.expect(try env.pop_local());

				defer env.discard(rhs);

				const lhs = try env.expect(try env.pop_local());

				defer env.discard(lhs);

				try env.locals.push_one(try env.new_boolean(try env.compare(lhs, rhs) > 0));
			},

			.clt => {
				const rhs = try env.expect(try env.pop_local());

				defer env.discard(rhs);

				const lhs = try env.expect(try env.pop_local());

				defer env.discard(lhs);

				try env.locals.push_one(try env.new_boolean(try env.compare(lhs, rhs) < 0));
			},

			.cge => {
				const rhs = try env.expect(try env.pop_local());

				defer env.discard(rhs);

				const lhs = try env.expect(try env.pop_local());

				defer env.discard(lhs);

				try env.locals.push_one(try env.new_boolean(try env.compare(lhs, rhs) >= 0));
			},

			.cle => {
				const rhs = try env.expect(try env.pop_local());

				defer env.discard(rhs);

				const lhs = try env.expect(try env.pop_local());

				defer env.discard(lhs);

				try env.locals.push_one(try env.new_boolean(try env.compare(lhs, rhs) <= 0));
			},

			.jf => |jf| {
				if (try env.pop_local()) |condition| {
					defer env.discard(condition);

					if (!condition.is_truthy()) {
						self.cursor = jf;
					}
				} else {
					self.cursor = jf;
				}
			},

			.jt => |jt| {
				if (try env.pop_local()) |condition| {
					defer env.discard(condition);

					if (condition.is_truthy()) {
						self.cursor = jt;
					}
				}
			},
		}
	}

	return env.pop_local();
}

pub fn free(self: *Self, env: *kym.RuntimeEnv) void {
	while (self.constants.pop()) |constant| {
		env.discard(constant);
	}

	self.constants.free();
	self.opcodes.free();
	env.discard(self.name);

	if (self.bindings.len != 0) {
		for (self.bindings) |binding| {
			if (binding) |value| {
				env.discard(value);
			}
		}

		env.allocator.deallocate(self.bindings);
	}

	self.bindings = &.{};
}

pub fn make(env: *kym.RuntimeEnv, name: []const coral.io.Byte, environment: *const tree.Environment) kym.RuntimeError!Self {
	var chunk = Self{
		.name = try env.new_symbol(name),
		.opcodes = OpcodeList.make(env.allocator),
		.constants = ConstList.make(env.allocator),
		.lines = LineList.make(env.allocator),
		.bindings = &.{},
		.arity = environment.argument_count,
		.cursor = 0,
	};

	var compiler = Compiler{
		.chunk = &chunk,
		.env = env,
	};

	try compiler.compile_environment(environment);

	return chunk;
}

fn syscall_import(env: *kym.RuntimeEnv, frame: *const kym.Frame) kym.RuntimeError!?*kym.RuntimeRef {
	return env.import(file.Path.from(&.{try env.unwrap_string(try frame.expect_arg(env, 0))}));
}

fn syscall_print(env: *kym.RuntimeEnv, frame: *const kym.Frame) kym.RuntimeError!?*kym.RuntimeRef {
	env.print(try env.unwrap_string(try frame.expect_arg(env, 0)));

	return null;
}

fn syscall_vec2(env: *kym.RuntimeEnv, frame: *const kym.Frame) kym.RuntimeError!?*kym.RuntimeRef {
	const x = @as(f32, @floatCast(try env.unwrap_float(try frame.expect_arg(env, 0))));

	if (frame.has_arg(env, 1)) |y| {
		return env.new_vector2(x, @floatCast(try env.unwrap_float(y)));
	}

	return env.new_vector2(x, x);
}

fn syscall_vec3(env: *kym.RuntimeEnv, frame: *const kym.Frame) kym.RuntimeError!?*kym.RuntimeRef {
	const x = @as(f32, @floatCast(try env.unwrap_float(try frame.expect_arg(env, 0))));

	if (frame.has_arg(env, 1)) |y| {
		return env.new_vector3(
			x,
			@floatCast(try env.unwrap_float(y)),
			@floatCast(try env.unwrap_float(try frame.expect_arg(env, 2))),
		);
	}

	return env.new_vector3(x, x, x);
}

pub const typeinfo = &kym.Typeinfo{
	.size = @sizeOf(Self),
	.name = "lambda",
	.destruct = typeinfo_destruct,
	.call = typeinfo_call,
	.to_string = typeinfo_to_string,
};

fn typeinfo_call(env: *kym.RuntimeEnv, userdata: []coral.io.Byte, frame: *const kym.Frame) kym.RuntimeError!?*kym.RuntimeRef {
	return @as(*Self, @ptrCast(@alignCast(userdata))).execute(env, frame);
}

fn typeinfo_destruct(env: *kym.RuntimeEnv, userdata: []coral.io.Byte) void {
	@as(*Self, @ptrCast(@alignCast(userdata))).free(env);
}

fn typeinfo_to_string(env: *kym.RuntimeEnv, userdata: []coral.io.Byte) coral.io.AllocationError!*kym.RuntimeRef {
	return env.to_string(@as(*Self, @ptrCast(@alignCast(userdata))).name);
}

pub fn write(self: *Self, line: u32, opcode: Opcode) coral.io.AllocationError!void {
	try self.opcodes.push_one(opcode);
	try self.lines.push_one(line);
}
